@@ -1,41 +1,49 @@
 import json
 import os
+import sys
 import time
 import uuid
 
 import httplib2
 from googleapiclient.discovery import build
 from googleapiclient.errors import HttpError
-from googleapiclient.http import MediaFileUpload
+from googleapiclient.http import MediaFileUpload, MediaIoBaseUpload
 from oauth2client.client import flow_from_clientsecrets
 from oauth2client.file import Storage
 from oauth2client.tools import run_flow
 
-from youtube_upload import (MAX_RETRIES, MISSING_CLIENT_SECRETS_MESSAGE,
-                            RETRYABLE_EXCEPTIONS, RETRYABLE_STATUS_CODES,
-                            VALID_PRIVACY_STATUSES, YOUTUBE_API_SERVICE_NAME,
-                            YOUTUBE_API_VERSION, YOUTUBE_UPLOAD_SCOPE)
+from youtube_upload import CLIENT_SECRETS_FILE
+from youtube_upload import MAX_RETRIES
+from youtube_upload import MISSING_CLIENT_SECRETS_MESSAGE
+from youtube_upload import OAUTH_FILE
+from youtube_upload import RETRYABLE_EXCEPTIONS
+from youtube_upload import RETRYABLE_STATUS_CODES
+from youtube_upload import VALID_PRIVACY_STATUSES
+from youtube_upload import YOUTUBE_API_SERVICE_NAME
+from youtube_upload import YOUTUBE_API_VERSION
+from youtube_upload import YOUTUBE_UPLOAD_SCOPE
+
 from youtube_upload.oauth_template import oauth_template
 
 
 class YoutubeUploader():
     '''
-    The YouTube Uploader client.
+    The YouTube Uploader service.
 
     When using in a multithreaded environment, please create a new instance of the `YoutubeUploader` class per thread.
     '''
 
     def __init__(
         self,
-        client_id="",
-        client_secret="",
+        client_id=None,
+        client_secret=None,
         secrets_file_path=os.path.join(
             '.',
-            'client_secrets.json')):
+            CLIENT_SECRETS_FILE)):
         '''
-        Initialization Function for the class. 
+        Initialization Function for the class.
 
-        The variables `client_id` and `client_secret` can be passed in when the class is initialized, this will have the function generate the `client_secrets.json` file. 
+        The variables `client_id` and `client_secret` can be passed in when the class is initialized, this will have the function generate the `client_secrets.json` file.
 
         If you do not pass the variables in, the class will look for a `client_secrets.json` file in the same direction in which the script is being initialized. You can instead
         pass in a directory to where the `client_secrets.json` is with the parameter `secrets_file_path` here is an example `client_secrets.json` file:
@@ -53,13 +61,10 @@ class YoutubeUploader():
         '''
 
         self.client_secrets = {}
-        if client_id == "" or client_secret == "":
+        if client_id is None or client_secret is None:
             self.secrets_file = secrets_file_path
-            try:
-                with open(secrets_file_path) as f:
-                    self.client_secrets = json.loads(f.read())
-            except FileNotFoundError:
-                raise
+            with open(secrets_file_path) as f:
+                self.client_secrets = json.loads(f.read())
         else:
             self.client_secrets = {
                 'web': {
@@ -78,7 +83,7 @@ class YoutubeUploader():
         self.options = None
         self.flow = None
         self.credentials = None
-        self.oauth_path = 'oauth.json'
+        self.oauth_path = OAUTH_FILE
 
         self.max_retry = MAX_RETRIES
 
@@ -86,7 +91,11 @@ class YoutubeUploader():
         self.close()
 
     # This is if you have another OAuth file to use
-    def authenticate(self, oauth_path='oauth.json', access_token=None, refresh_token=None):
+    def authenticate(
+            self,
+            oauth_path=OAUTH_FILE,
+            access_token=None,
+            refresh_token=None):
         '''
         This method authenticates the user with Google's servers. If you give no path, the method will look for the `oauth.json` file in the current working directory.
 
@@ -109,8 +118,7 @@ class YoutubeUploader():
                 'access_token': access_token,
                 'refresh_token': refresh_token,
                 'client_id': self.client_secrets['web'].get('client_id'),
-                'client_secret': self.client_secrets['web'].get('client_secret')
-            }
+                'client_secret': self.client_secrets['web'].get('client_secret')}
             oauth_json_str = oauth_template.substitute(subs)
             with open(self.oauth_path, 'w') as f:
                 f.write(oauth_json_str)
@@ -133,9 +141,9 @@ class YoutubeUploader():
             http=self.credentials.authorize(
                 httplib2.Http()))
 
-    def upload(self, file_path, options={}, chunksize=(-1)):
+    def upload(self, file_path, options=None, chunksize=(-1)):
         '''
-        This uploads the file to YouTube. The only required argument is the `file_path`, which is the path to the video to be uploaded. 
+        This uploads the file to YouTube. The only required argument is the `file_path`, which is the path to the video to be uploaded.
 
         The `options` parameter is a dictionary of options. The items are pretty self explanatory, here is an example options dictionary:
         ```Python
@@ -151,31 +159,18 @@ class YoutubeUploader():
         }
         ```
 
-        The parameter, `chunk_size` is the max size of the HTTP request to send the video. This parameter is in bytes, and if set to `-1`, which is the default, it 
-        will send the video in one large request. Set this to a different value if you are having issues with the upload failing. 
+        The parameter, `chunk_size` is the max size of the HTTP request to send the video. This parameter is in bytes, and if set to `-1`, which is the default, it
+        will send the video in one large request. Set this to a different value if you are having issues with the upload failing.
 
-        You can also add a callback function for the case where the upload were to fail. This is mainly for custom error handling and prompting users to re-authenticate. 
-
-        The callback function is the parameter `noauth_callback`. The function passed should be able to accept a tuple of arguments. The parameter `noauth_args` is the arguments for the function. Here 
-        is an example of the callback function being used.
+        Will return the response from YouTube, as well as the response of the thumbnail upload as a tuple.
 
         ```Python
-        from youtube_upload.client import YoutubeUploader
-
-        def callback(*args):
-            print(args[0], args[1])
-
-        with YoutubeUploader() as youtube:
-            youtube.authenticate()
-            youtube.upload("test.mp4", noauth_callback=callback, noauth_args=("It failed."))
+        response, thumbnail_response = client.upload(file_path, options)
         ```
 
-        And here is the example output in the case that the upload would fail.
-
-        ```
-        It failed. invalid_grant: Bad Request
-        ```
         '''
+        if options is None:
+            options = {}
         body = {
             'snippet': {
                 'title': options.get('title', 'Test Title'),
@@ -195,34 +190,115 @@ class YoutubeUploader():
                     body.keys())), body=body, media_body=MediaFileUpload(
                 file_path, chunksize=chunksize, resumable=True))
 
-        self._resumable_upload(
+        return self._resumable_upload(
+            insert_request, bool(
+                options.get('thumbnailLink')), options)
+
+    def upload_stream(self, file_object, options=None, chunksize=(-1)):
+        '''
+        Uploads the file to YouTube from the specified file-like object.
+
+        We are using this to stream files from S3 to YouTube!
+
+        ```python
+        import s3fs
+        from youtube_upload.client import YouTubeUploader
+        fs = s3fs.S3FileSystem(anon=True)
+
+        video = fs.open('s3://bucket/video.mp4')
+
+        client = YouTubeUploader()
+
+        client.authenticate()
+
+        client.upload_stream(video)
+
+        client.close()
+        video.close()
+
+        ```
+
+        The `options` parameter is a dictionary of options. The items are pretty self explanatory, here is an example options dictionary:
+        ```Python
+        # Video options
+        options = {
+            title : "Example title",
+            description : "Example description",
+            tags : ["tag1", "tag2", "tag3"],
+            categoryId : "22",
+            privacyStatus : "private",
+            kids : False
+            thumbnailLink : "https://cdn.havecamerawilltravel.com/photographer/files/2020/01/youtube-logo-new-1068x510.jpg"
+        }
+        ```
+
+        The parameter, `chunk_size` is the max size of the HTTP request to send the video. This parameter is in bytes, and if set to `-1`, which is the default, it
+        will send the video in one large request. Set this to a different value if you are having issues with the upload failing.
+
+        Will return the response from YouTube, as well as the response of the thumbnail upload as a tuple.
+
+        ```Python
+        response, thumbnail_response = client.upload(file_path, options)
+        ```
+        '''
+        if options is None:
+            options = {}
+        body = {
+            'snippet': {
+                'title': options.get('title', 'Test Title'),
+                'description': options.get('description', 'Test Description'),
+                'tags': options.get('tags'),
+                'categoryId': options.get('category', '22')
+            },
+            'status': {
+                'privacyStatus': options.get('privacyStatus', VALID_PRIVACY_STATUSES[0]),
+                'selfDeclaredMadeForKids': options.get('kids', False)
+            }
+        }
+
+        media = MediaIoBaseUpload(
+            file_object,
+            "application/octet-stream",
+            chunksize=chunksize,
+            resumable=True)
+
+        insert_request = self.youtube.videos().insert(
+            part=",".join(
+                list(
+                    body.keys())), body=body, media_body=media)
+
+        return self._resumable_upload(
             insert_request, bool(
                 options.get('thumbnailLink')), options)
 
     def _resumable_upload(self, insert_request, uploadThumbnail, options):
         response = None
+        thumbnail_response = None
         error = None
         retry = 0
 
         while response is None:
             try:
                 _, response = insert_request.next_chunk()
+                #skipcq: PYL-R1723
                 if 'id' in response:
                     video_id = response.get('id')
                     if uploadThumbnail:
                         request = self.youtube.thumbnails().set(
                             videoId=video_id, media_body=MediaFileUpload(
                                 options.get('thumbnailLink')))
-                        response = request.execute()
-                        print(response)
+                        thumbnail_response = request.execute()
+                    break
+
                 else:
-                    raise HttpError(f'Unexpected response: {response}')
+                    # skipcq: PYL-E1120
+                    raise Exception(f'Unexpected response: {response}')
             except HttpError as e:
                 if e.resp.status in RETRYABLE_STATUS_CODES:
                     error = "A retryable HTTP error %d occurred:\n%s" % (
                         e.resp.status, e.content)
                 else:
-                    raise
+                    raise e
             except RETRYABLE_EXCEPTIONS as e:
                 error = "A retryable error occurred: %s" % e
 
@@ -230,11 +306,14 @@ class YoutubeUploader():
                 print(error)
                 retry += 1
                 if retry > self.max_retry:
-                    exit("No longer attempting to retry.")
+                    #skipcq: PYL-E1120
+                    raise Exception('Exceeded max retries. ' + error)
 
                 print("Sleeping 5 seconds and then retrying...")
                 time.sleep(5)
 
+        return response, thumbnail_response
+
     def close(self):
         '''
         Tears down and closes the class cleanly.
